/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <getopt.h>

#define PROGRAM_NAME "rac"

#define VERSION "1.0"

#define AUTHORS "Aurelio Marinho Jargas, Thobias Salazar Trevisan"

#define BUG_ADDRESS "verde@aurelio.net, thobias@thobias.org"

/* matrix allocate memory (step) */
#define LINE_SIZE 100
#define LINE_NUM  200

#define TYPE_RANGE_LINE 0
#define TYPE_RANGE_COL  1

/* command allocate memory (step) */
#define CMD_SIZE 15

#ifndef LONG_MAX
#define LONG_MAX   2147483647L
#endif

char **line;
long int num_line=1;
long int *num_chars_line;
long int line_start, line_end;
char *col_start, *col_end;
FILE *stream;
long int STEP_LINE;
long int STEP_COL;
int TYPE;

void usage(int status)
{
	if (status !=0)
		fprintf (stderr, "Tente `%s --help' para mais informaes.\n",PROGRAM_NAME);
	else {
		fprintf(stdout,"\
Uso: %s ENDEREO [ARQUIVO]\n\
 ou: %s OPES\n\
\n\
OPES\n\
\n\
     -h, --help        mostra estas informaes e sai\n\
     -V, --version     mostra a verso do programa e sai\n\
\n\
ENDEREO\n\
\n\
     Linha   Coluna    Significado\n\
     -----------------------------------------------\n\
       N       [N]     N\n\
     N;M;I   [N,M,I]   N, M e I\n\
      N:M     [N:M]    trecho de N at M\n\
      N~M     [N~M]     partir de N, ande de M em M\n\
\n\
Se o ARQUIVO for '-', l o texto da entrada padro (STDIN).\n",PROGRAM_NAME,PROGRAM_NAME);
fprintf(stdout,"Se o nmero do endereo for negativo, conta  partir do fim.\n\
Envie relatrios de erros para: %s\n",BUG_ADDRESS);
	}
	exit (status == 0 ? 0 : 1);
}

void get_file()
{
	int c;
	long int i=0;
	long int allocated_num_line=LINE_NUM;
	long int actual_line_size=0;

	/* alloc vector to number of chars per line */
	if (!(num_chars_line = (long int *) malloc(LINE_NUM * sizeof(long int)))){
		fprintf(stderr, "Error: alloc vector to number of chars per line\n");
			exit(1);
	}
	/* alloc numer of lines' memory */
	if (!(line = (char **) malloc(LINE_NUM * sizeof(char *)))){
		fprintf(stderr, "Error: alloc number of lines\n");
		exit(1);
	}

	while (1) {
		i++;
		c = getc(stream);
		/* size of line is enough ?! */
		if (actual_line_size - 1 <= i) {
			if (!actual_line_size) line[num_line] = NULL;
#ifdef DEBUG2
			printf("-- alloc size of line. line number = %li, size line = %li\n",num_line,actual_line_size);
#endif
			actual_line_size += LINE_SIZE;
			if (actual_line_size >= LONG_MAX) {
				fprintf (stderr, "Error: overflow actual_line_size\n");
				exit(1);
			}
			line[num_line] = (char *) realloc (line[num_line], actual_line_size * sizeof (char));
			if (!line[num_line]) {
					fprintf (stderr, "Error: realloc line's memory\n");
					exit(1);
			}
		}
		line[num_line][i] = c;
		if (c == '\n' || c == EOF || c == '\r') {
			/* is this a DOS file ??*/
			if (c == '\r') 	{ 
				c = getc(stream);
				if (c != '\n') {
						fprintf(stderr,"Error: unknown file\n");
						exit(1);
				}
			}
			line[num_line][i+1] = '\0';
			i--;
			if (c == EOF) 
					line[num_line][i+1] = '\0';
			num_chars_line[num_line] = i;

			if (c == EOF) {
				if (i == 0) num_line--;
				break;
			}
			else {
				i = 0;
				actual_line_size = 0;
				/* dont need check. we check the allocated_num_line */
				num_line++;
				/* number of lines are enough ?! */
				if (num_line >= allocated_num_line - 1) {
#ifdef DEBUG2
					printf("-- realloc number of lines. line number = %li\n",num_line);
#endif
					allocated_num_line += LINE_NUM;
					if (allocated_num_line >= LONG_MAX) {
						fprintf (stderr, "Error: overflow allocated_num_line\n");
						exit(1);
					}
					line = (char **) realloc (line, allocated_num_line * sizeof (char *));
					if (!line) {
						fprintf (stderr, "Error: realloc num line's memory\n");
						exit(1);
					}
					num_chars_line = (long int *) realloc(num_chars_line, allocated_num_line * sizeof(long int));
					if (!num_chars_line) {
						fprintf(stderr, "Error: realloc vector to number of chars per line\n");
						exit(1);
					}
				}
			}
		}
	}
}

/*
 * TODO: Merge the check_col and check_lines functions.
 *       Just one check and filter!!!
 * */
void check_col(long int *line, long int *col)
{
	if (*col < 0) 
		*col = num_chars_line[*line] + (*col+1);

	if (*col < 0) *col = 1;
	if (*col > num_chars_line[*line]) *col = 0;
}

int check_lines()
{
	if (line_start > num_line && line_end < 0) return -1;	
	if (line_start > num_line && line_end >= num_line) return -1;	
	
	/* check. we cant overflow the limits */
	if (line_start < 0) line_start = num_line + (line_start+1);
	if (line_end   < 0) line_end = num_line + (line_end+1);

	if (line_start == line_end)
			if (line_end < 1 || line_end > num_line)
					return -1;

	if (line_start >= 1 && line_start <= num_line) {
			if (line_end < 1) line_end = 1;
			if (line_end > num_line) line_end = num_line;
	}
		
	if (line_end <= num_line && line_end >= 1) {
			if (line_start < 1) line_start = 1;
			if (line_start > num_line) line_start = num_line;
	}

#ifdef DEBUG1
	printf("-- Definitive Range = from %li to %li\n",line_start,line_end);
#endif
	
	/* to guarantee that we do not overflow the limits of matrix */
	if (line_start < 1 || line_start > num_line) return -1;
	if (line_end < 1 || line_end > num_line) return -1;

	return 0;
}

void get_type(char *command)
{
	char c;
	int i=0;

	for (i=0;i<=strlen(command);i++) {
		c = command[i];
		if (c == ':')
			TYPE = TYPE_RANGE_LINE;
		else if (c == '[') {
			while (c != ']' && c){
				i++;
				c = command[i];
			}
		}
	}
	if (TYPE == -1) TYPE = 1;
}

/* We know. This function is really horrible. 
 * TODO: please improve!!!  */
void exec_range_lines(char *command)
{
	long int i, j, start, end;
	
	if (line_start < 0 && line_end < 0)
		line_end = num_line + (line_end+1);
	if (line_start > 0 && line_end + num_line < 0) return;
	if (line_start < 0 && line_end > num_line) return;
	if (line_start + num_line < 0 && line_end <= 1) return;
	if (check_lines() != 0) return;

	if (line_start <= line_end) {
		for (i=line_start;i<=line_end;i=i+STEP_LINE) {
			start = col_start[0] == '\0' ? 1 : atoi(col_start); 
			end = col_end[0] == '\0' ? -1 : atoi(col_end);

			check_col(&i,&start);
			check_col(&i,&end);

			if (i == line_start) {
				for (j=start;j<=num_chars_line[i];j++)
					if (start != 0)
						printf("%c",line[i][j]);
			}
			else if (i == line_end) {
				if (end == 0) end = num_chars_line[i];
					for (j=1;j<=end;j++)
						printf("%c",line[i][j]);
			}
			else {
				for (j=1;j<=num_chars_line[i];j+=STEP_LINE)
					printf("%c",line[i][j]);
			}
			printf("\n");
		}
	}
	else {
		STEP_LINE = -1;
		for (i=line_start;i>=line_end;i=i+STEP_LINE) { 
			start = col_start[0] == '\0' ? 1 : atoi(col_start); 
			end = col_end[0] == '\0' ? -1 : atoi(col_end);

			check_col(&i,&start);
			check_col(&i,&end);
			if (i == line_start) {
				for (j=start;j<=num_chars_line[i];j++)
					printf("%c",line[i][j]);
			}
			else if (i == line_end) {
				for (j=1;j<=end;j++)
					printf("%c",line[i][j]);
			}
			else {
				for (j=1;j<=num_chars_line[i];j++)
					printf("%c",line[i][j]);
			}
			printf("\n");
		}
	}
}

/*
 * TODO: This function too.
 *       needs improvements 
 */
void exec_range_cols()
{
	long int i, j, start, end;
	int all_line;
	char *temp, *token, *col_temp1, *col_temp2, *col_STEP;

	if (col_start[0] == '\0' && col_end[0] != '\0')
			strncpy(col_start,col_end,strlen(col_end));

	if (line_start + num_line < 0 && line_end <= 1 && STEP_LINE < 0) return;
	if (check_lines() != 0) return;

	i = line_start;

	if (line_start > line_end) {
			j = line_start;
			line_start = line_end;
			line_end = j;
	}
	
	while (i >= line_start && i <= line_end) {
		temp = (char *) malloc(strlen(col_start)+1);
		temp[0]='\0';
		all_line = 0;
		strcpy(temp,col_start);
		token = strtok(temp, ",");
		STEP_COL=1;
		while (token != NULL) {
			all_line = 1;
			col_temp1 = (char *) malloc (strlen(token)+1);
			col_temp1[0] = '\0';
			strncpy(col_temp1,token,strlen(token));
			token = strtok(NULL, ",");
			col_temp2 = strchr(col_temp1,':');
			col_STEP  = strchr(col_temp1,'~');
			/* is it a range ? */
			if (col_temp2 != NULL || col_STEP != NULL) {
				if (col_temp2 != NULL) {
					if (col_temp1[0] == ':') start = 1;
					else start = atoi(col_temp1);
				
					*col_temp2 = 0;
					col_temp2++;
			
					if (atoi(col_temp2) == 0)
						end = -1;
					else {
						end = atoi(col_temp2);
						if (start < 0 && end > num_chars_line[i]) 
								continue;
						if (end > num_chars_line[i])
							end = -1;
					}
				}	
				else if (col_STEP != NULL) {
						*col_STEP=0;
						col_STEP++;
						
						STEP_COL = atoi(col_STEP);
						
						if (col_temp1[0] == ':') start = 1;
						else start = atoi(col_temp1);


						if (STEP_COL > 0) end = -1;
						else if (STEP_COL < 0) end = 1;
						else {
							fprintf(stderr,"Error: step col invalid\n");
							exit(1);
						}
				}
				if (start > num_chars_line[i] && 
							num_chars_line[i]+STEP_COL < 0) continue;

				if (start > num_chars_line[i] && end <= 0) continue;
				
				if (num_chars_line[i] + start < 0 && STEP_COL <=0)
						continue;
			
				if (start > 0 && end+num_chars_line[i] < 0) continue;

				if (start+num_chars_line[i] < 0 && end > 0) continue;
				
				check_col(&i,&start);
				check_col(&i,&end);
				if (start == 0) start = num_chars_line[i];
				if (start == 1 && end <= 0) continue;

#ifdef DEBUG1
				printf("-- col start %li col end = %li\n",start,end);
#endif
				if (start <= end) {
					for (j=start;j<=end;j=j+STEP_COL) {
						if (j < 0 ) goto out; 
						if (j > 0 && j <= num_chars_line[i])
							printf("%c",line[i][j]);
					}
				}
				else {
					if (STEP_COL == 1) STEP_COL = -1;
					for (j=start;j>=end;j=j+STEP_COL)
						if (j > 0 && j <= num_chars_line[i])
							printf("%c",line[i][j]);
				}
			}
			else { 
				start = atoi(col_temp1);
				if (start < 0)
					start = start + num_chars_line[i] + 1;
				if (start > 0 && start <= num_chars_line[i])
					printf("%c",line[i][start]);
			}
out:
			free(col_temp1);
		}
		/* print line*/
		if (all_line == 0) {
			for (j=1;j<=num_chars_line[i];j++)
				printf("%c",line[i][j]);
		}
		printf("\n");
		i = i + STEP_LINE;
	}
}

void exec_cmd(char *command)
{

#ifdef DEBUG1
	printf("-- col_start = %s\n",col_start);
	printf("-- col_end   = %s\n",col_end);
#endif

	if (TYPE == TYPE_RANGE_LINE) {
#ifdef DEBUG1
		printf("-- TYPE = TYPE_RANGE_LINE\n");
#endif
		exec_range_lines(command);
	}
	
	if (TYPE == TYPE_RANGE_COL) {
#ifdef DEBUG1
		printf("-- TYPE = TYPE_RANGE_COL\n");
#endif
		exec_range_cols();
	}
}


void get_ranges(char *command)
{
	char *p1, c;

	p1 = (char *) malloc(strlen(command)+1);
	free(col_start);
	free(col_end);
	col_start = (char *) malloc(strlen(command)+1);
	col_end   = (char *) malloc(strlen(command)+1);
	p1[0] = '\0';
	col_start[0] = '\0';
	col_end[0]   = '\0';
	strncpy(&c,command,1);
	while(c) {
		if (c == ':' || c == '~') {
			if (atoi(p1) == 0) 
				line_start = 1;
			else 
				line_start = atoi(p1);
			if (c == '~') {
				command++;
				strncpy(&c,command,1);
				STEP_LINE = atoi(command);
				if (STEP_LINE >= 0)line_end = -1;
				else line_end = 1;
				while (c != '[' && c) {
					command++;
					strncpy(&c,command,1);
				}
			}
			else
				command++;
			strncpy(&c,command,1);
			*p1 = 0;
			while(c) {
				if (c == '[') {
					while(c != ']' && c) {
						if (c != '[' && c != ']') 
							strncat(col_end,command,1);
						command++;
						strncpy(&c,command,1);
					}
					if (c != ']') {
						fprintf(stderr, "Error: Invalid Range\n");
						exit(1);
					}
#ifdef DEBUG2
					printf("-- col_end = %s\n",col_end);
#endif
				}
				strncat(p1,command,1);
				command++;
				strncpy(&c,command,1);
			}
			if (line_end != -1 && line_end != 1) {
				if (atoi(p1) == 0) line_end = num_line;
				else line_end = atoi(p1);
			}
		}
		if (c == '[') {
			line_start = atoi(p1);
			while(c != ']' && c) {
				if (c != '[' && c != ']') strncat(col_start,command,1);
				command++;
				strncpy(&c,command,1);
			}
			if (c != ']') {
				fprintf(stderr, "Error: Invalid Range\n");
				exit(1);
			}
#ifdef DEBUG2
			printf("-- col_start = %s\n",col_start);
#endif
		}
		strncat(p1,command,1);
		command++;
		strncpy(&c,command,1);
	}

	if (col_start[0] == '\0' && line_start == 0) {
			line_start = atoi(p1);
			if (col_end[0] == '\0' && line_end == 0) 
				line_end = atoi(p1);
	}
	if (col_start[0] != '\0' && line_start != 0 && col_end[0] == '\0'  && line_end == 0) 
			line_end = line_start;
	if (col_start[0] != '\0' && line_start == 0 && col_end[0] == '\0'  && line_end == 0) {
			line_start = 1;
			line_end   = num_line;
	}
	free(p1);
#ifdef DEBUG1
	printf("-- line_start = %li\n",line_start);
	printf("-- line_end   = %li\n",line_end);
#endif
}


int main(int argc, char *argv[])
{
	extern int opterr;
	long int i, j, k;
	int opt, cmd_num, cmd_size;
	char *token, **command;

	static struct option longopts[] = {
		{"version", 0, NULL, 'V'},
		{"help", 0, NULL, 'h'},
		{NULL, 0, NULL, 0}
	};

	/* to suppress error message generated by getopt() */
	opterr = 0;
	while((opt = getopt_long(argc, argv, "hV", longopts, NULL)) != -1) {
		switch(opt) {
			case 'h':
				usage(0);
			case 'V':
				 fprintf(stdout, "%s verso %s\n", PROGRAM_NAME,VERSION);
				 exit(0);
		}
	}

	if (argc < 2)
		usage(1);

	if (argc == 2 || (!strcmp(argv[2],"-"))) {
#ifdef DEBUG1
		printf("-- Reading from stdin\n");
#endif
		stream = stdin;
	}
	else {
		if (!(stream = fopen(argv[2], "r"))) {
			perror(argv[2]);
			exit(1);
		}
	}

	/* put file into matrix */
	get_file();

#ifdef DEBUG1
	printf("-- Total number of lines %li\n",num_line);
#endif

	if (!(command = (char **) malloc(CMD_SIZE * sizeof (char *)))){
		fprintf(stderr, "Error: alloc command's memory\n");
		exit(1);
	}

	cmd_num = CMD_SIZE;
	cmd_size = 0;
	
	/* Get command by command */
	token = strtok(argv[1], ";");
	k=0;
	while (token != NULL) {
		for (i=0, j=0;i<strlen(token);i++) {
			if (cmd_size == 0) 	command[k] = NULL;
			/* erase the blank characters */
			if (j >= cmd_size - 1) {
#ifdef DEBUG2
				printf("-- line = %li, realloc size of commands = %d\n",k,cmd_size);
#endif
				cmd_size += CMD_SIZE;
				if (cmd_size >= INT_MAX) {
					fprintf (stderr, "Error: overflow cmd_size\n");
					exit(1);
				}
				command[k] = (char *) realloc (command[k], cmd_size * sizeof (char));
				if (!command[k]) {
					fprintf (stderr, "Error: realloc command's memory\n");
					exit(1);
				}
			}
			if (token[i] != ' ') command[k][j++] = token[i];
		}
		token = strtok(NULL, ";");
		command[k][j] = '\0';
		cmd_size = 0;
		k++;
		if (k >= cmd_num) {
#ifdef DEBUG2
			printf("-- realloc number of commands = %d\n",cmd_num);
#endif
			cmd_num += CMD_SIZE;
			if (cmd_num >= INT_MAX) {
				fprintf (stderr, "Error: overflow cmd_num\n");
				exit(1);
			}
			command = (char **) realloc(command, cmd_num * sizeof(char *));
			if (!command) {
				fprintf (stderr, "Error: realloc command's memory\n");
				exit(1);
			}
		}
	}
	for (i=0;i<k;i++) {
#ifdef DEBUG1
		printf("-- Command = %s\n",command[i]);
#endif
		line_start = line_end = 0;
		STEP_LINE=1;
		TYPE=-1;
		/* it executes the command */
		get_ranges(command[i]);
		get_type(command[i]);
		exec_cmd(command[i]);
		free(command[i]);
	}

	for (i=0;i<num_line;i++) free(line[i]);
	free(line);
	fclose(stream);

	exit(0);
}

